#include <snort.h>

/*
* Snort parsing is a 3 step process.
*
* 1. Build a list of key/value pairs (for our matchers)
* 2. Optimise/fix that list of matchers, eg: make 'msg' data
*    become the argument for the target etc..
* 3. Actually build and add the rule(s), then cleanup the key/val
*    list.
*/

proc_print_out print_out;
proc_print_raw print_raw;

proc_add_rule		add_rule;
proc_add_criteria	add_criteria;
proc_set_target		set_target;
proc_add_parser		add_parser;
proc_rule_cleanup	rule_cleanup;

proc_list_append	list_append;
proc_list_remove	list_remove;

/* This global data isn't protected. that means
* this plugin is in NO WAY reentrant. */
char *target=NULL;
char *t_priv=NULL;
p_snort temp=NULL;

int snort_add_rule(void)
{
	p_snort p;
	p_rule r;
	
	char *real_t_priv;
	
	if ( !t_priv )
		real_t_priv="SNORT";
	else
		real_t_priv=t_priv;
	
	if ( !(r=malloc(sizeof(rule_t))) ) return 0;
	SAFE_INIT_RULE_PTR(r);

	/* Set the rules target */
	if ( !(set_target(r, target, real_t_priv)) ) {
		rule_cleanup(r);
		return 0;
	}
	
	for (p=temp; p; p=(p_snort)p->l.next) {
		if ( !add_criteria(r, p->key, p->val) ) {
			rule_cleanup(r);
			return 0;
		}
	}
	
	if ( !add_rule(r) ) {
		rule_cleanup(r);
		return 0;
	}
	
	return 1;
}

void snort_cleanup_internal(void)
{
	p_snort foo=NULL, bar=NULL;
	
	if ( target ) free(target);
	if ( t_priv ) free(t_priv);

	for (foo=temp; foo;) {
		if ( foo->key ) free(foo->key);
		if ( foo->val ) free(foo->val);
		bar=foo;
		foo=foo->l.next;
		free(bar);
	}
	temp=NULL;
}

int snort_link_element(const char *key, const char *val)
{
	p_snort n=NULL;
	
	if ( !key || !val ) return 0;

	if ( !(n=malloc(sizeof(snort_t))) ) return 0;

	if ( !(n->key=strdup(key)) ) {
		free(n);
		return 0;
	}
	
	if ( !(n->val=strdup(val)) ) {
		free(n->key);
		free(n);
		return 0;
	}

	temp=list_append(n, temp);
	return 1;
}

int snort_optimise_rule(p_snort tokens)
{
	p_snort p,t;

	/* Turn the MSG param in to the private data */
	for (p=temp; p; ) {
		if ( !strcasecmp(p->key, "MSG" ) ) {
			if ( !p->l.prev ) {
				temp=(p_snort)p->l.next;
				t=list_remove(p);
			}else{
				t=list_remove(p);
			}
			
			p=(p_snort)p->l.next;
			t_priv=t->val;
			free(t->key);
			free(t);
			continue;
		}
		p=(p_snort)p->l.next;
	}
	
	return 1;
}

/* TODO:
* Allow the escaping of quote marks inside quoted strings
* Strip the quote marks off of quoted strings
* */
#define PARSE_CRAP	0
#define PARSE_KEY	1
#define PARSE_VAL	2
#define PARSE_END	3
int snort_parse_body(char *body)
{
	char *cur=NULL;
	int  cur_meaning=PARSE_CRAP;
	char *begin_key=NULL;
	char *begin_val=NULL;
	int keys=0;
	int quote=0;
	char *end;

	end=body+strlen(body);

	for ( cur=body; *cur <= end; cur++ ) {
		switch ( cur_meaning ) {
			case PARSE_CRAP:
				if ( quote ) break;
				if ( *cur == '(' ) break;
				if ( *cur == ')' ) break;
				if ( !isblank(*cur) ) {
					cur_meaning=PARSE_KEY;
					begin_key=cur;
				}
				break;
			case PARSE_KEY:
				if ( quote ) break;
				if ( *cur == ':' ) {
					*cur='\0';
					cur_meaning=PARSE_VAL;
					begin_val=cur+1;
				}
				break;
			case PARSE_VAL:
				if ( *cur == '"' ) {
					quote=(quote) ? 0 : 1;
				}
				if ( quote ) break;
				if ( *cur == ';' ) {
					*cur='\0';
					cur_meaning=PARSE_END;
					
				}
				break;
			case PARSE_END:
				snort_link_element(begin_key, begin_val);
				keys++;
				cur_meaning=PARSE_CRAP;
				begin_val=NULL;
				begin_key=NULL;
				break;
		}
	}
	
	return keys ? 1 : 0;
}

snort_ret snort_parse_rule(char *rule)
{
	p_rule nrule;
	char *tokens[5];
	int i=0;
	char *work;
	snort_ret ret=SNORT_NULL;
	
	target=strdup(strtok(rule," \t"));
	if ( !target ) return SNORT_NULL;
	
	/* Get the head part tokenised:
	* target src spt -> dst dpt */
	while ( (tokens[i]=strtok(NULL," \t"))  && (i++<5) );
	if ( i != 6 ) return SNORT_NULL;
	
	/* Which way the rule goes, back, forth, bi */
	if ( !strcmp(tokens[3], "<-") ) { ret=SNORT_BACKWARDS; }
	if ( !strcmp(tokens[3], "->") ) { ret=SNORT_FORWARDS; }
	if ( !strcmp(tokens[3], "<>") ) { ret=SNORT_BIDIRECTIONAL; }
	
	/* Has to go back OR forth */
	if ( ret == SNORT_NULL ) return ret;
	
	/* Add the head stuff to the rule */
	snort_link_element("IP_PROTO", tokens[0]);
	if ( strcasecmp("any", tokens[1]) ) snort_link_element("IP_SRC", tokens[1]);
	if ( strcasecmp("any", tokens[2]) ) snort_link_element("SPT", tokens[2]);
	if ( strcasecmp("any", tokens[4]) ) snort_link_element("IP_DST", tokens[4]);
	if ( strcasecmp("any", tokens[5]) ) snort_link_element("DPT", tokens[5]);
	

	/* We now have the variable sized rule body
	* in this format:
	* (key:123; key2:"val"; key:"val |0d|";)  */
	
	if ( !(work=strdup(strtok(NULL, ""))) ) {
		return SNORT_NULL;
	}
	
	if ( !snort_parse_body(work) ) ret = SNORT_NULL;
	
	free(work);
	return ret;
}

int snort_loadfile(char *fn)
{
	FILE *f;
	char buf[4096];
	char *p=(char *)&buf;
	int rule=0, line=0;
	snort_ret ret;
	
	if ( !fn ) return 0;
	
	if ( (f=fopen(fn, "r")) == NULL ) {
		return 0;
	}

	print_out("Opening snort rulefile (%s)" CRLF, fn);

	while( fgets(p, 4096, f) ) {
		line++;
		if ( IS_RULE(buf) ) {
			rule++;
			ret = snort_parse_rule(p);
			if ( ret == SNORT_NULL ) {
				print_out("Parse error on rule %i (%s:%i)" CRLF,
					rule, fn, line);
			}else{
				if ( snort_optimise_rule(temp) )
					snort_add_rule();
			}

			/* Clean up our parsers little scratch area */
			snort_cleanup_internal();
		}
	}
	
	fclose(f);
	return 1;
}

uint _PLUGIN_INIT (p_plugin_in in, p_plugin_out out)
{
	PLUGIN_CHECK_PARAMS(in);

	/* Take the functions we need */
	print_out=in->print_out;
	print_raw=in->print_raw;
	add_rule=in->add_rule;
	add_criteria=in->add_criteria;
	set_target=in->set_target;
	add_parser=in->add_parser;
	rule_cleanup=in->rule_cleanup;
	list_append=in->list_append;
	list_remove=in->list_remove;
	
	/* Fill in a description about ourselves */
	out->name="Snort";
	out->desc="Loads snort ruleset files";
	out->ver_major=1;
	out->ver_minor=0;
	
	/* Tell firestorm what function to
	* call when it needs to load a 'Snort' file */
	add_parser("Snort", snort_loadfile);
	
	return PLUGIN_ERR_OK;
}

/* This is where we would cleanup our shit */
uint _PLUGIN_UNLOAD (void)
{
	return PLUGIN_ERR_OK;
}
