#include <matcher.h>
#include <s_util.h>
#include <cleanup.h>
#include <targets.h>

/* List of registered matchers   */
p_matcher	matches=NULL;
mlock_t		matches_lock;

/* List of all added rules */
p_rule		rules;
mlock_t		rules_lock;

int matcher_add_rule(p_rule r)
{
	if ( r == NULL ) return 0;

	/* FIXME: validate rule better */
	if ( r->criteria == NULL ) return 0;

	CRITICAL_START(matches_lock);
	rules=list_append(r, rules);
	CRITICAL_STOP(matches_lock);

	return 1;
}

int matcher_set_rule_target(p_rule r, char *target, char *val)
{
	p_target t;
	
	if ( !r || !target ) return 0;

	if ( !(t=target_get(target)) ) return 0;

	if ( !(r->t_priv=t->validate(val)) ) return 0;

	r->target=t;
	
	return 1;
}

/* Add a criteria to a rule */
int matcher_add_criteria_to_rule(p_rule r, char *crit, char *val)
{
	p_matcher matcher=NULL;
	p_relement rel=NULL;
	
	if ( !r || !crit || !val ) return 0;

	if ( (matcher=matcher_get(crit)) == NULL ) {
		print_out(PRINT_ERROR "Unable to add rule, no such "
			"matcher (%s)" CRLF, crit);
		return 0;
	}

	if ( (rel=malloc(sizeof(relement_t))) == NULL  ) {
		return 0;
	}

	/* Validate this rule criteria */
	rel->match = matcher;
	if ( (rel->priv=rel->match->validate(val)) == NULL ) {
		free(rel);
		return 0;
	}

	/* Now add the criteria to the rule */
	r->criteria = list_append(rel, r->criteria);
	return 1;
}

/* 
* Cleans up a rule object (MUST have been
* safely initialised with SAFE_INIT_RULE_PTR() 
* before calling this
*/
void matcher_rule_cleanup(p_rule r)
{
	p_relement r_foo, r_bar;

	if ( r->target && r->t_priv )
		r->target->cleanup(r->t_priv);
	
	for (r_foo=r->criteria; r_foo;) {
		r_foo->match->cleanup(r_foo->priv);
			
		r_bar=r_foo;
		r_foo=(p_relement)r_foo->l.next;
		free(r_bar);
	}

	free(r);
}

void matcher_cleanup(int code, void *priv)
{
	p_rule foo,bar;
	p_matcher m_foo, m_bar;
	
	print_raw("Removing all rules" CRLF);
	CRITICAL_START(rules_lock);
	for (foo=rules; foo;) {
		bar=foo;
		foo=(p_rule)foo->l.next;
		matcher_rule_cleanup(bar);
	}
	CRITICAL_STOP(rules_lock);
	
	print_raw("Removing all matchers" CRLF);
	CRITICAL_START(matches_lock);
	for (m_foo=matches; m_foo;) {
		m_bar=m_foo;
		m_foo=(p_matcher)m_foo->l.next;
		free(m_bar);
	}
	CRITICAL_STOP(matches_lock);
}

int matcher_init(void)
{
	MUTEX_CREATE(matches_lock, "Matches");
	cleanup_add_handler(cleanup_lock, (void *)&matches_lock);

	MUTEX_CREATE(rules_lock, "Rules");
	cleanup_add_handler(cleanup_lock, (void *)&rules_lock);

	cleanup_add_handler(matcher_cleanup, NULL);
	
	return 1;
}

/* Get a matcher by its name  */
p_matcher matcher_get(char *n)
{
	p_matcher tmp;

	for ( tmp=matches; tmp; tmp=(p_matcher)tmp->l.next ) {
		if ( strcasecmp(tmp->name, n) == 0 ) {
			return tmp;
		}
	}

	return NULL;
}

/* Plugins call this to add a new matcher */
int matcher_add ( char *n, proc_match_validate v, 
		proc_match_cleanup cl,
		proc_match_match match )
{
	p_matcher foo;

	if ( n == NULL ) return 0;
	if ( v == NULL ) return 0;
	if ( cl == NULL ) return 0;
	if ( match == NULL ) return 0;
	
	if ( matcher_get(n) ) {
		print_out(PRINT_WARNING "Tried to load %s matcher twice" CRLF, n);
		return 0;
	}
	
	if ( !(foo=malloc(sizeof(matcher_t))) ) {
		return 0;
	}

	foo->name=n;
	foo->validate=v;
	foo->cleanup=cl;
	foo->match=match;
	
	CRITICAL_START(matches_lock);
	matches=list_append(foo, matches);
	CRITICAL_STOP(matches_lock);

	return 1;
}

/* Actually match a packet agains all our rules */
void matcher_begin( p_packet p )
{	
	int m=1;
	p_rule r=NULL;
	p_relement c=NULL;
	
	for (r=rules; r; r=(p_rule)r->l.next) {
		m=1;
		
		for (c=r->criteria; c; c=(p_relement)c->l.next) {
			if ( !(c->match->match(c->priv, p)) ){
				m=0;
				break;
			}
		}
		
		if (m) {
			/* Send the packet to the rule target */
			r->target->target(p, r->t_priv);
			return;
		}
	}

	//print_raw(PRINT_ERROR "Packet not matched!" CRLF);
}
